"""
Setup repro environment.

Downloads and installs particular mongodb versions (each binary is renamed
to include its version) into an install directory and symlinks the binaries
with versions to another directory. This script supports community and
enterprise builds.
"""
import argparse
import logging
import os
import re
import sys
from typing import List, NamedTuple, Optional

import inject
import structlog
from evergreen import EvergreenApi

from db_contrib_tool.clients.download_client import DownloadError
from db_contrib_tool.clients.file_service import FileService
from db_contrib_tool.clients.resmoke_proxy import ResmokeProxy
from db_contrib_tool.config import (
    SETUP_REPRO_ENV_CONFIG,
    SETUP_REPRO_ENV_CONFIG_FILE,
    WINDOWS_BIN_PATHS_FILE,
    DownloadTarget,
    SetupReproEnvConfig,
)
from db_contrib_tool.plugin import PluginInterface, Subcommand, SubcommandResult
from db_contrib_tool.services.evergreen_service import EvergreenService
from db_contrib_tool.setup_repro_env.artifact_discovery_service import (
    ArtifactDiscoveryService,
    RequestTarget,
    RequestType,
)
from db_contrib_tool.setup_repro_env.download_service import (
    ArtifactDownloadService,
    DownloadOptions,
)
from db_contrib_tool.utils import evergreen_conn, is_windows

SUBCOMMAND = "setup-repro-env"
BINARY_ARTIFACT_NAME = "Binaries"
KNOWN_BRANCHES = {"master"}
RELEASE_VERSION_RE = re.compile(r"^\d+\.\d+$")
PATCH_VERSION_RE = re.compile(r"^\d+\.\d+\.\d+")
BRANCH_RE = re.compile(r"^v\d+\.\d+")
EXTERNAL_LOGGERS = [
    "evergreen",
    "github",
    "inject",
    "segment",
    "urllib3",
]

LOGGER = structlog.getLogger(__name__)


class SetupReproEnvError(Exception):
    """Errors in setup_repro_env.py.

    The base class of exceptions for this file/subcommand.
    """

    pass


def setup_logging(debug=False):
    """Enable logging."""
    log_level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(
        format="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
        level=log_level,
        stream=sys.stdout,
    )
    for logger in EXTERNAL_LOGGERS:
        logging.getLogger(logger).setLevel(logging.WARNING)
    structlog.configure(logger_factory=structlog.stdlib.LoggerFactory())


class SetupReproParameters(NamedTuple):
    """
    Parameters describing how a repro environment should be setup.

    * edition: MongoDB edition to download.
    * platform: Target platform to download.
    * architecture: Target architecture to download.
    * variant: Build Variant to download from.

    * versions: List of items to download.
    * install_last_lts: If True download last LTS version of mongo.
    * install_last_continuous: If True download last continuous version of mongo.
    * ignore_failed_push: Download version even if the push task failed.
    * fallback_to_master: Should the latest master be downloaded if the version doesn't exist.

    * evg_version_file: Write which evergreen version were downloaded from to this file.

    * download_options: Options specifying how downloads should occur.
    """

    edition: str
    platform: str
    architecture: str
    variant: str

    versions: List[str]
    install_last_lts: bool
    install_last_continuous: bool
    ignore_failed_push: bool
    fallback_to_master: bool

    evg_version_file: Optional[str]

    download_options: DownloadOptions

    def get_download_target(self, platform: Optional[str] = None) -> DownloadTarget:
        """
        Get the download target to use based on these parameters.

        :param platform: Override the platform with this platform.
        :return: Download target specified by this options.
        """
        platform = platform if platform is not None else self.platform
        return DownloadTarget(
            edition=self.edition, platform=platform, architecture=self.architecture
        )


class SetupReproOrchestrator:
    """Orchestrator for setting up repro environments."""

    @inject.autoparams()
    def __init__(
        self,
        evg_service: EvergreenService,
        resmoke_proxy: ResmokeProxy,
        artifact_download_service: ArtifactDownloadService,
        artifact_discovery_service: ArtifactDiscoveryService,
        file_service: FileService,
    ) -> None:
        """
        Initialize the orchestrator.

        :param evg_service: Service for working with evergreen.
        :param resmoke_proxy: Proxy for working with resmoke.
        :param artifact_download_service: Service to download artifacts.
        :param artifact_discovery_service: Service to find artifacts.
        :param file_service: Service to work with the filesystem.
        """
        self.evg_service = evg_service
        self.resmoke_proxy = resmoke_proxy
        self.artifact_download_service = artifact_download_service
        self.artifact_discovery_service = artifact_discovery_service
        self.file_service = file_service

    def interpret_request(self, request: str) -> RequestTarget:
        """
        Translate the request from the user into an item we can understand.

        :param request: Request from user.
        :return: Targeted request to download.
        """
        if request in KNOWN_BRANCHES or BRANCH_RE.match(request):
            return RequestTarget(RequestType.GIT_BRANCH, request)

        if RELEASE_VERSION_RE.match(request):
            return RequestTarget(RequestType.MONGO_RELEASE_VERSION, request)

        if PATCH_VERSION_RE.match(request):
            return RequestTarget(RequestType.MONGO_PATCH_VERSION, request)

        if self.evg_service.query_task_existence(request):
            return RequestTarget(RequestType.EVG_TASK, request)

        if self.evg_service.query_version_existence(request):
            return RequestTarget(RequestType.EVG_VERSION, request)

        return RequestTarget(RequestType.GIT_COMMIT, request)

    def interpret_requests(
        self, request_list: List[str], last_lts: bool, last_continuous: bool
    ) -> List[RequestTarget]:
        """
        Translate all the requests from the user into items we can understand.

        :param request_list: Requests from user.
        :param last_lts: Should 'last lts' version be included.
        :param last_continuous: Should the 'last continuous' version be included.
        :return: List of targeted request to download.
        """
        requests = [self.interpret_request(request) for request in request_list]
        if last_lts or last_continuous:
            requests.extend(self._get_release_versions(last_lts, last_continuous))

        return requests

    def _get_release_versions(
        self, install_last_lts: Optional[bool], install_last_continuous: Optional[bool]
    ) -> List[RequestTarget]:
        """
        Create a list of multiversion versions that should be included.

        :param install_last_lts: True if the last LTS version should be included.
        :param install_last_continuous: True if the last continuous version should be included.
        :return: List of which multiversion versions should be included.
        """
        multiversionconstants = self.resmoke_proxy.get_multiversion_constants()
        releases = {
            multiversionconstants.last_lts_fcv: install_last_lts,
            multiversionconstants.last_continuous_fcv: install_last_continuous,
        }
        LOGGER.debug("LTS and continuous release inclusions", releases=releases)
        out = {
            RequestTarget.previous_release(version)
            for version, requested in releases.items()
            if requested
        }

        return list(out)

    @staticmethod
    def _get_bin_suffix(version: str, evg_project_id: str) -> str:
        """Get the multiversion bin suffix from the evergreen project ID."""
        if re.match(r"(\d+\.\d+)", version):
            # If the cmdline version is already a semvar, just use that.
            return version
        elif evg_project_id in ("mongodb-mongo-master", "mongodb-mongo-master-nightly"):
            # If the version is not a semvar and the project is the master waterfall,
            # we can't add a suffix.
            return ""
        else:
            # Use the Evergreen project ID as fallback.
            return re.search(r"(\d+\.\d+$)", evg_project_id).group(0)

    def execute(self, setup_repro_params: SetupReproParameters) -> bool:
        """Execute setup repro env mongodb."""
        request_list = self.interpret_requests(
            setup_repro_params.versions,
            setup_repro_params.install_last_lts,
            setup_repro_params.install_last_continuous,
        )

        downloaded_versions = []
        failed_requests = []
        link_directories = []

        download_target = setup_repro_params.get_download_target()
        LOGGER.info("Search criteria", search_criteria=download_target)

        for request in request_list:
            LOGGER.info("Setting up request", request=request)
            LOGGER.info("Fetching download URLs from Evergreen")

            try:
                urls_info = self.artifact_discovery_service.find_artifacts(
                    request,
                    setup_repro_params.variant,
                    download_target,
                    setup_repro_params.ignore_failed_push,
                    setup_repro_params.fallback_to_master,
                )

                if urls_info is None:
                    failed_requests.append(request)
                    LOGGER.warning("Unable to find artifacts for request", request=request)
                    continue

                bin_suffix = self._get_bin_suffix(request.identifier, urls_info.project_identifier)
                linked_dir = self.artifact_download_service.download_and_extract(
                    urls_info.urls,
                    bin_suffix,
                    urls_info.evg_version_id,
                    setup_repro_params.download_options,
                )
                if linked_dir:
                    link_directories.append(linked_dir)
                downloaded_versions.append(urls_info.evg_version_id)
                LOGGER.info("Setup request completed", request=request)
            except (
                evergreen_conn.EvergreenConnError,
                DownloadError,
                SetupReproEnvError,
            ):
                failed_requests.append(request)
                LOGGER.error("Setup request failed", request=request, exc_info=True)

        if is_windows():
            self.file_service.write_windows_install_paths(WINDOWS_BIN_PATHS_FILE, link_directories)

        if setup_repro_params.evg_version_file is not None:
            self.file_service.append_lines_to_file(
                setup_repro_params.evg_version_file, downloaded_versions
            )
            LOGGER.info(
                "Finished writing downloaded Evergreen versions",
                target_file=os.path.abspath(setup_repro_params.evg_version_file),
            )

        if len(downloaded_versions) < len(request_list):
            LOGGER.error("Some requests were not able to setup.", failed_requests=failed_requests)
            return False
        LOGGER.info("Downloaded versions", request_list=request_list)
        return True


class SetupReproEnv(Subcommand):
    """Main class for the setup repro environment subcommand."""

    def __init__(
        self,
        download_options,
        install_dir="",
        link_dir="",
        mv_platform=None,
        edition=None,
        architecture=None,
        versions=None,
        variant=None,
        install_last_lts=None,
        install_last_continuous=None,
        evergreen_config=None,
        debug=None,
        ignore_failed_push=False,
        evg_versions_file=None,
        resmoke_cmd=None,
        fallback_to_master=False,
    ):
        """Initialize."""
        setup_logging(debug)

        download_options = DownloadOptions(
            download_binaries=download_options.download_binaries,
            download_symbols=download_options.download_symbols,
            download_artifacts=download_options.download_artifacts,
            download_python_venv=download_options.download_python_venv,
            install_dir=os.path.abspath(install_dir),
            link_dir=os.path.abspath(link_dir),
        )

        self.setup_repro_params = SetupReproParameters(
            edition=edition.lower() if edition else None,
            platform=mv_platform.lower() if mv_platform else None,
            architecture=architecture.lower() if architecture else None,
            variant=variant.lower() if variant else None,
            versions=versions,
            install_last_lts=install_last_lts,
            install_last_continuous=install_last_continuous,
            ignore_failed_push=ignore_failed_push,
            fallback_to_master=fallback_to_master,
            download_options=download_options,
            evg_version_file=evg_versions_file,
        )

        self.evg_api = evergreen_conn.get_evergreen_api(evergreen_config)
        self.resmoke_cmd = resmoke_cmd

    def execute(self):
        """Execute setup repro env mongodb."""

        def dependencies(binder: inject.Binder) -> None:
            """Define dependencies for execution."""
            binder.bind(SetupReproEnvConfig, SETUP_REPRO_ENV_CONFIG)
            binder.bind(EvergreenApi, self.evg_api)
            binder.bind(ResmokeProxy, ResmokeProxy.with_cmd(self.resmoke_cmd))

        inject.configure(dependencies)

        setup_repro_orchestrator = inject.instance(SetupReproOrchestrator)

        success = setup_repro_orchestrator.execute(self.setup_repro_params)
        if success:
            return SubcommandResult.SUCCESS
        return SubcommandResult.FAIL


class _DownloadOptions(object):
    def __init__(self, db, ds, da, dv):
        self.download_binaries = db
        self.download_symbols = ds
        self.download_artifacts = da
        self.download_python_venv = dv


class SetupReproEnvPlugin(PluginInterface):
    """Integration point for setup-repro-env."""

    DEFAULT_INSTALL_DIR = os.path.join(os.getcwd(), "build", "multiversion_bin")
    DEFAULT_LINK_DIR = os.getcwd()
    DEFAULT_WITH_ARTIFACTS_INSTALL_DIR = os.path.join(os.getcwd(), "repro_envs")
    DEFAULT_WITH_ARTIFACTS_LINK_DIR = os.path.join(
        DEFAULT_WITH_ARTIFACTS_INSTALL_DIR, "multiversion_bin"
    )

    @classmethod
    def _update_args(cls, args):
        """Update command-line arguments."""
        if not args.versions:
            args.install_last_lts = True
            args.install_last_continuous = True

        if args.download_artifacts:
            args.install_dir = cls.DEFAULT_WITH_ARTIFACTS_INSTALL_DIR
            args.link_dir = cls.DEFAULT_WITH_ARTIFACTS_LINK_DIR

    def parse(self, subcommand, parser, parsed_args, **kwargs):
        """Parse command-line arguments."""
        if subcommand != SUBCOMMAND:
            return None

        # Shorthand for brevity.
        args = parsed_args
        self._update_args(args)

        download_options = _DownloadOptions(
            db=(not args.skip_binaries),
            ds=args.download_symbols,
            da=args.download_artifacts,
            dv=args.download_python_venv,
        )

        if download_options.download_binaries and args.link_dir is None:
            raise ValueError("link_dir must be specified if downloading binaries")

        return SetupReproEnv(
            install_dir=args.install_dir,
            link_dir=args.link_dir,
            mv_platform=args.platform,
            edition=args.edition,
            architecture=args.architecture,
            versions=args.versions,
            install_last_lts=args.install_last_lts,
            variant=args.variant,
            install_last_continuous=args.install_last_continuous,
            download_options=download_options,
            evergreen_config=args.evergreen_config,
            ignore_failed_push=(not args.require_push),
            evg_versions_file=args.evg_versions_file,
            debug=args.debug,
            resmoke_cmd=args.resmoke_cmd,
            fallback_to_master=args.fallback_to_master,
        )

    @classmethod
    def _add_args_to_parser(cls, parser):
        parser.add_argument(
            "-i",
            "--installDir",
            dest="install_dir",
            default=cls.DEFAULT_INSTALL_DIR,
            help=f"Directory to install the download archive,"
            f" [default: %(default)s, if `--downloadArtifacts` is passed: {cls.DEFAULT_WITH_ARTIFACTS_INSTALL_DIR}]",
        )
        parser.add_argument(
            "-l",
            "--linkDir",
            dest="link_dir",
            default=cls.DEFAULT_LINK_DIR,
            help=f"Directory to contain links to all binaries for each version in the install directory,"
            f" [default: %(default)s, if `--downloadArtifacts` is passed: {cls.DEFAULT_WITH_ARTIFACTS_LINK_DIR}]",
        )
        editions = ("base", "enterprise", "targeted")
        parser.add_argument(
            "-e",
            "--edition",
            dest="edition",
            choices=editions,
            default="enterprise",
            help="Edition of the build to download, [default: %(default)s].",
        )
        parser.add_argument(
            "-p",
            "--platform",
            dest="platform",
            help="Platform to download. "
            f"Available platforms can be found in {SETUP_REPRO_ENV_CONFIG_FILE}.",
        )
        parser.add_argument(
            "-a",
            "--architecture",
            dest="architecture",
            default="x86_64",
            help="Architecture to download, [default: %(default)s]. Examples include: "
            "'arm64', 'ppc64le', 's390x' and 'x86_64'.",
        )
        parser.add_argument(
            "-v",
            "--variant",
            dest="variant",
            default=None,
            help="Specify a variant to use, which supersedes the --platform, --edition and"
            " --architecture options.",
        )
        parser.add_argument(
            "versions",
            nargs="*",
            help="Accepts binary versions, `master`, full git commit hashes, evergreen version ids,"
            " evergreen task ids. Binary version examples: <major.minor>, 4.2, 4.4, 5.0 etc. If no"
            " version is specified the last LTS and the last continuous versions will be installed.",
        )
        parser.add_argument(
            "--installLastLTS",
            dest="install_last_lts",
            action="store_true",
            help="If specified, the last LTS version will be installed",
        )
        parser.add_argument(
            "--installLastContinuous",
            dest="install_last_continuous",
            action="store_true",
            help="If specified, the last continuous version will be installed",
        )
        parser.add_argument(
            "-sb",
            "--skipBinaries",
            dest="skip_binaries",
            action="store_true",
            help="whether to skip downloading binaries.",
        )
        parser.add_argument(
            "-ds",
            "--downloadSymbols",
            dest="download_symbols",
            action="store_true",
            help="whether to download debug symbols.",
        )
        parser.add_argument(
            "-da",
            "--downloadArtifacts",
            dest="download_artifacts",
            action="store_true",
            help="whether to download artifacts.",
        )
        parser.add_argument(
            "-dv",
            "--downloadPythonVenv",
            dest="download_python_venv",
            action="store_true",
            help="whether to download python venv.",
        )
        parser.add_argument(
            "-ec",
            "--evergreenConfig",
            dest="evergreen_config",
            help="Location of evergreen configuration file. If not specified it will look "
            f"for it in the following locations: {evergreen_conn.EVERGREEN_CONFIG_LOCATIONS}",
        )
        parser.add_argument(
            "-d",
            "--debug",
            dest="debug",
            action="store_true",
            help="Set DEBUG logging level.",
        )
        parser.add_argument(
            "-rp",
            "--require-push",
            dest="require_push",
            action="store_true",
            help="Require the push task to be successful for assets to be downloaded",
        )
        parser.add_argument(
            "--resmokeCmd",
            dest="resmoke_cmd",
            default="python buildscripts/resmoke.py",
            help="Command to invoke resmoke.py",
        )
        parser.add_argument(
            "--fallbackToMaster",
            dest="fallback_to_master",
            action="store_true",
            help=(
                "Fallback to downloading the latest binaries from master if the requested "
                "version is not found (Only application for mongo versions)"
            ),
        )
        # Hidden flag to write out the Evergreen versions of the downloaded binaries.
        parser.add_argument(
            "--evgVersionsFile", dest="evg_versions_file", default=None, help=argparse.SUPPRESS
        )

    def add_subcommand(self, subparsers):
        """Create and add the parser for the subcommand."""
        parser = subparsers.add_parser(SUBCOMMAND, help=__doc__)
        self._add_args_to_parser(parser)



"""Command-line entry-point into db-contrib-tool."""

import argparse
import sys

from db_contrib_tool.evg_aware_bisect import BisectPlugin
from db_contrib_tool.setup_repro_env.setup_repro_env import SetupReproEnvPlugin
from db_contrib_tool.symbolizer.mongosymb import SymbolizerPlugin
from db_contrib_tool.usage_analytics import track_usage

_PLUGINS = [BisectPlugin(), SetupReproEnvPlugin(), SymbolizerPlugin()]


def parse(sys_args):
    """Parse the CLI args."""

    parser = argparse.ArgumentParser(
        description="The db-contrib-tool - MongoDB's tools for contributors."
        " For more information, see the help message for each subcommand."
        " For example: db-contrib-tool setup-repro-env -h"
    )
    subparsers = parser.add_subparsers(dest="command")

    # Add sub-commands.
    for plugin in _PLUGINS:
        plugin.add_subcommand(subparsers)

    parsed_args = parser.parse_args(sys_args)

    return parser, parsed_args


def parse_command_line(sys_args, **kwargs):
    """Parse the command line arguments passed to db-contrib-tool and return the subcommand object to execute."""
    parser, parsed_args = parse(sys_args)

    command = parsed_args.command

    track_usage(
        event="Command Line Tool Invoked",
        properties={
            "command": command,
            "parsed_arguments": vars(parsed_args),
        },
    )

    for plugin in _PLUGINS:
        subcommand_obj = plugin.parse(command, parser, parsed_args, **kwargs)
        if subcommand_obj is not None:
            return subcommand_obj


def main():
    subcommand = parse_command_line(sys.argv[1:])
    result = subcommand.execute()
    sys.exit(result.value)


if __name__ == "__main__":
    main()



"""Interactions with the db-contrib-tool bisect command."""
import logging
import os
import subprocess
import sys
from datetime import datetime, timedelta, timezone

import structlog

from db_contrib_tool.config import SetupReproEnvConfig
from db_contrib_tool.plugin import PluginInterface, Subcommand, SubcommandResult
from db_contrib_tool.services.evergreen_service import EvergreenService
from db_contrib_tool.utils import evergreen_conn

LOGGER = structlog.getLogger(__name__)

_HELP = """
Perform an evergreen-aware git-bisect to find the 'last passing version' and 'first failing version'
of mongo, with respect to a user provided shell script.
"""

_USAGE = """
The 'bisect' command lets a user specify a '--branch', '--variant' & '--lookback' period on which to
perform a bisect. The user also provides a shell test '--script' which exits with status code 0 to
indicate a successful test. The command performs the following steps:

(1) Get all versions for the given '--branch', '--variant' & '--lookback' period from Evergreen.
(2) Filter the versions for versions that Evergreen has binaries and artifacts for.
(3) Find the 'middle' version.
(4) Setup a test environment.
    - The 'build/resmoke-bisect' directory will have a sub directory --
    'build/resmoke-bisect/{version_id}' containing the git repo for this version.
    - The 'binaries' & 'artifacts' will also be downloaded to the directory named
    'build/resmoke-bisect/{version_id}'.
    - Create a virtual environment at 'build/resmoke-bisect/bisect_venv' and
    install packages for this version.
(5) Activate 'bisect_venv' & run the user provided shell script from within the
'build/resmoke-bisect/{version_id}' directory.
(6) Teardown the test environment.
(7) Repeat steps (3)-(6) on the left half, if (5) failed, or right half, if (5) succeeded.

This command will print the "Last Known Passing Version" & "First Known Failing Version".

NOTE: This 'bisect' command assumes a perfect partition between passing & failing versions.
ie: [Pass, Pass, Pass, Fail, Fail, Fail]
If there is not a perfect partition, try modifying the '--lookback' period or shell '--script'.
"""

_COMMAND = "bisect"

BISECT_DIR = os.path.dirname(os.path.abspath(__file__))
SETUP_TEST_ENV_SH = os.path.join(BISECT_DIR, "setup_test_env.sh")
TEARDOWN_TEST_ENV_SH = os.path.join(BISECT_DIR, "teardown_test_env.sh")
RUN_USER_SCRIPT_SH = os.path.join(BISECT_DIR, "run_user_script.sh")


def setup_logging(debug=False):
    """Enable logging."""
    log_level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(
        format="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
        level=log_level,
        stream=sys.stdout,
    )
    structlog.configure(logger_factory=structlog.stdlib.LoggerFactory())


class Bisect(Subcommand):
    """Main class for the Bisect subcommand."""

    def __init__(
        self,
        branch,
        lookback,
        variant,
        script,
        python_installation,
        evergreen_config=None,
        debug=None,
    ):
        """Initialize."""
        setup_logging(debug)
        self.branch = branch
        self.lookback = lookback
        self.variant = variant
        self.script = script
        self.python_installation = python_installation
        self.evergreen_config = evergreen_config
        self.evg_api = evergreen_conn.get_evergreen_api(evergreen_config)
        self.evg_service = EvergreenService(self.evg_api, SetupReproEnvConfig.empty_config())

    @staticmethod
    def _teardown_test_env(version):
        """Remove any directories added during testing process."""
        try:
            subprocess.run(["bash", TEARDOWN_TEST_ENV_SH], check=True)
        except subprocess.CalledProcessError as err:
            LOGGER.error("Could not teardown test environment for bisect.", version=version)
            raise err
        LOGGER.info("Completed teardown of test environment for bisect.", version=version)

    def _setup_test_env(self, version):
        """Set up a test environment for the given version."""
        try:
            subprocess.run(
                [
                    "bash",
                    SETUP_TEST_ENV_SH,
                    self.python_installation,
                    self.evergreen_config if self.evergreen_config else "",
                    self.variant,
                    version,
                ],
                check=True,
            )
        except subprocess.CalledProcessError:
            LOGGER.error(
                "Could not setup test environment for bisect -- retrying.", version=version
            )
            try:
                subprocess.run(
                    [
                        "bash",
                        SETUP_TEST_ENV_SH,
                        self.python_installation,
                        self.evergreen_config if self.evergreen_config else "",
                        self.variant,
                        version,
                    ],
                    check=True,
                )
            except subprocess.CalledProcessError as err:
                LOGGER.error("Could not setup test environment for bisect.", version=version)
                raise err
        LOGGER.info("Completed setup of test environment for bisect.", version=version)

    def _run_user_script(self, version):
        """Run the user script in a virtual environment."""
        return subprocess.run(
            ["bash", RUN_USER_SCRIPT_SH, version, self.script], check=False
        ).returncode

    def _test_version_with_script(self, version):
        """Test the given version with the user provided script."""
        self._setup_test_env(version)
        success = self._run_user_script(version)
        self._teardown_test_env(version)
        return success == 0

    def bisect(self, versions):
        """Bisect to find latest passing version assuming ordered by oldest to latest version."""
        if len(versions) == 0:
            return None
        midpoint = len(versions) // 2
        success = self._test_version_with_script(versions[midpoint])
        # if success, keep checking right
        if success:
            LOGGER.info(
                "Version passed user script.",
                version=versions[midpoint],
            )
            return self.bisect(versions[midpoint + 1 :]) or versions[midpoint]
        # if fail, keep checking left
        else:
            LOGGER.info("Version failed user script.", version=versions[midpoint])
            return self.bisect(versions[0:midpoint])

    def find_versions_with_binaries(self):
        """
        Find versions that have binaries for the user provided variant in the lookback period.

        :return: List of versions that have binaries ordered from earliest to latest.
        """
        versions_with_binaries = []
        for version in self.evg_api.versions_by_project_time_window(
            project_id=f'mongodb-mongo-{"" if self.branch == "master" else "v"}{self.branch}',
            before=datetime.now(timezone.utc),
            after=datetime.now(timezone.utc) - timedelta(self.lookback),
        ):
            urls = self.evg_service.get_compile_artifact_urls(
                version, self.variant, ignore_failed_push=True
            )
            if urls is not None and urls.urls.get("Artifacts") and urls.urls.get("Binaries"):
                versions_with_binaries.append(version.revision)

        versions_with_binaries.reverse()
        return versions_with_binaries

    def execute(self):
        """
        Perform bisect for the provided branch, variant & lookback period based on result of script.

        Print the last passing version and first failing version.
        """
        versions_with_binaries = self.find_versions_with_binaries()

        # No versions found
        if not versions_with_binaries:
            LOGGER.info(
                "No versions with binaries found for given branch, variant & lookback period.",
                branch=self.branch,
                variant=self.variant,
                lookback=self.lookback,
            )
            return SubcommandResult.SUCCESS

        LOGGER.info(
            "Performing bisect on the following versions: ", versions=versions_with_binaries
        )

        last_passing_version = self.bisect(versions_with_binaries)
        first_failing_version = None

        # All versions failed
        if last_passing_version is None:
            first_failing_version = versions_with_binaries[0]
            LOGGER.info("All versions in lookback period failed.")

        # All versions passed
        elif last_passing_version == versions_with_binaries[-1]:
            LOGGER.info("All versions in lookback period passed.")

        else:
            first_failing_version = versions_with_binaries[
                versions_with_binaries.index(last_passing_version) + 1
            ]

        print(f"Last Known Passing Version: {last_passing_version}")
        print(f"First Known Failing Version: {first_failing_version}")
        return SubcommandResult.SUCCESS


class BisectPlugin(PluginInterface):
    """Interact with Resmoke bisect."""

    TOOLCHAIN_PYTHON3 = "/opt/mongodbtoolchain/v3/bin/python3"

    @classmethod
    def _add_args_to_parser(cls, parser):
        parser.add_argument(
            "--lookback",
            "-l",
            action="store",
            type=int,
            default=365,
            help="Maximum number of days to look back for versions to test.",
        )
        parser.add_argument(
            "--branch",
            "-b",
            action="store",
            type=str,
            required=True,
            help="The branch for which versions are being tested. [REQUIRED]",
        )
        parser.add_argument(
            "--variant",
            "-v",
            action="store",
            type=str,
            required=True,
            help="The variant for which versions are being tested. [REQUIRED]",
        )
        parser.add_argument(
            "--script",
            "-s",
            action="store",
            type=str,
            required=True,
            help="Location of the shell test script to run on the versions. [REQUIRED]",
        )
        parser.add_argument(
            "--python-installation",
            "-p",
            action="store",
            type=str,
            dest="python_installation",
            default="python3",
            required=False,
            help="Location of a python installation to use for shell commands. If not specified, "
            f"'python3' or '{cls.TOOLCHAIN_PYTHON3}' (if toolchain is setup) will be used.",
        )
        parser.add_argument(
            "-ec",
            "--evergreenConfig",
            dest="evergreen_config",
            help="Location of evergreen configuration file. If not specified it will look "
            f"for it in the following locations: {evergreen_conn.EVERGREEN_CONFIG_LOCATIONS}",
        )
        parser.add_argument(
            "-d",
            "--debug",
            dest="debug",
            action="store_true",
            default=False,
            help="Set DEBUG logging level.",
        )

    def add_subcommand(self, subparsers):
        """
        Add 'bisect' subcommand.

        :param subparsers: argparse parser to add to.
        """
        parser = subparsers.add_parser(_COMMAND, usage=_USAGE, help=_HELP)
        self._add_args_to_parser(parser)

    @classmethod
    def _update_args(cls, args):
        """Update command-line arguments."""
        if os.path.isfile(cls.TOOLCHAIN_PYTHON3):
            args.python_installation = cls.TOOLCHAIN_PYTHON3

    def parse(self, subcommand, parser, parsed_args, **kwargs):
        """
        Return bisect if command is one we recognize.

        :param subcommand: equivalent to parsed_args.command
        :param parser: parser used
        :param parsed_args: output of parsing
        :param kwargs: additional args
        :return: None or a Subcommand
        """
        if subcommand != _COMMAND:
            return None

        args = parsed_args
        self._update_args(args)

        return Bisect(
            branch=args.branch,
            lookback=args.lookback,
            evergreen_config=args.evergreen_config,
            variant=args.variant,
            script=args.script,
            debug=args.debug,
            python_installation=args.python_installation,
        )



"""Service to parse backtraces in mongo log files."""
import sys
from json import JSONDecodeError, JSONDecoder
from typing import Any, Dict, Optional, Union

import inject


class MongoLogParserService:
    """Service to parse backtraces in mongo log files."""

    @inject.autoparams()
    def __init__(self, json_decoder: JSONDecoder):
        """
        Initialize.

        :param json_decoder: JSONDecoder object.
        """
        self.json_decoder = json_decoder

    def parse_log_lines(self, log_lines: str) -> Optional[Dict[str, Any]]:
        """
        Analyze log lines until backtrace is found.

        :param log_lines: Log lines as string.
        :return: Backtrace dict or None.
        """
        for line in log_lines.splitlines():
            start_index = line.find("{")
            if start_index == -1:
                continue

            # Skip over everything before the first '{' since it is likely to be log line prefixes.
            sub_line = line[start_index:]
            try:
                # Using raw_decode() to ignore extra data after the closing '}' to allow maximal
                # sloppiness in copy-pasting input.
                possible_trace_doc = self.json_decoder.raw_decode(sub_line)[0]

                trace_doc = self.find_backtrace(possible_trace_doc)
                if trace_doc is not None:
                    return trace_doc

            except JSONDecodeError:
                # It is expected that any kind of logs may be passed in along with a backtrace,
                # we are skipping the lines that do not contain a valid JSON document
                pass

        return None

    def find_backtrace(self, log_line: Any) -> Optional[Dict[str, Any]]:
        """
        Search the log line recursively for a dict that has "backtrace" and "processInfo" keys.

        :param log_line: Log line as dict or part of the log line as any type.
        :return: Backtrace dict or None.
        """
        try:
            if "backtrace" in log_line.keys() and "processInfo" in log_line.keys():
                return log_line
            for val in log_line.values():
                res = self.find_backtrace(val)
                if res:
                    return res
        except AttributeError:
            pass
        return None



"""Interface for creating a db-contrib-tool plugin."""

import abc
from enum import Enum


class SubcommandResult(Enum):
    """Subcommand result."""

    SUCCESS = 0
    FAIL = 1


class Subcommand(abc.ABC):
    """A db-contrib-tool subcommand to execute."""

    def execute(self) -> SubcommandResult:
        """Execute the subcommand."""
        raise NotImplementedError()


class PluginInterface(abc.ABC):
    """Subcommand/plugin interface."""

    def add_subcommand(self, subparsers):
        """
        Add parser options for this plugin.

        :param subparsers: argparse subparsers
        """
        raise NotImplementedError()

    def parse(self, subcommand, parser, parsed_args, **kwargs):
        """
        Resolve command-line options to a Subcommand or None.

        :param subcommand: equivalent to parsed_args.command
        :param parser: parser used
        :param parsed_args: output of parsing
        :param kwargs: additional args
        :return: None or a Subcommand
        """
        raise NotImplementedError()
